#!/usr/bin/env python3
# Author: Anthony Ruhier

import logging

logging.basicConfig(
    format="[%(levelname)s] %(message)s (%(filename)s:%(lineno)d) "
)

from pyqos.app import PyQoS
from pyqos import algorithms



#!/usr/bin/env python3
# Author: Anthony Ruhier
# Declares decorators

from functools import wraps


def multiple_interfaces(f):
    """
    Handle multiple interfaces for tc

    If the parameter "interface" is a list of multiple interfaces, it will
    execute the function f for each interface
    """
    @wraps(f)
    def repeat_for_each_interface(interface=None, *args, **kwargs):
        if type(interface) is not str and interface is not None:
            for i in interface:
                repeat_for_each_interface(i, *args, **kwargs)
        else:
            # keep interface as optional attribute for some functions
            if interface is None:
                f(*args, **kwargs)
            else:
                f(interface, *args, **kwargs)
    return repeat_for_each_interface



#!/usr/bin/env python3
# Author: Anthony Ruhier

import argparse
import logging
import os
import subprocess
import sys

from pyqos.backend import tc
from pyqos.config import Config, ConfigAttribute

global_logger = logging.getLogger("pyqos")
_logger = logging.getLogger(__name__)


class PyQoS():
    """
    Application to simplify the initialization of the QoS rules. Inspired from
    the Flask project.

    Usually you create a :class:`PyQoS` instance in your main module or
    in the :file:`__init__.py` file of your package like this::
        from pyqos import PyQoS
        app = PyQoS(application_name)
    """
    #: set the main logger in debug mode or not
    debug = ConfigAttribute("DEBUG")
    #: dryrun
    dryrun = ConfigAttribute("DRYRUN")
    #: name of the main logger
    logger_name = ConfigAttribute('LOGGER_NAME')
    #: configuration default values
    default_config = {
        "DEBUG": False,
        "DRYRUN": False,
        "LOGGER_NAME": None,
        "INTERFACES": dict(),
    }

    #: list of qos object to apply at run
    run_list = list()

    def __init__(self, app_name="pyqos", root_path=None):
        self.app_name = app_name
        self.config = Config(root_path, self.default_config)
        self._logger = None
        self.logger_name = self.app_name

    @property
    def logger(self):
        """
        A :class:`logging.Logger` object for this application.  The
        default configuration is to log to stderr if the application is
        in debug mode.  This logger can be used to (surprise) log messages.
        Here some examples::
            app.logger.debug('A value for debugging')
            app.logger.warning('A warning occurred (%d apples)', 42)
            app.logger.error('An error occurred')
        """
        if not (self._logger and self._logger.name == self.logger_name):
            self._logger = logging.Logger(self.logger_name)
        if self.config["DEBUG"]:
            self._logger.setLevel(logging.DEBUG)
        else:
            self._logger.setLevel(logging.WARNING)
        return self._logger

    def get_ifnames(self, interfaces_lst=None):
        if interfaces_lst is None:
            interfaces_lst = self.config["INTERFACES"]
        if_names = set()
        for interface in interfaces_lst.values():
            if "name" in interface.keys():
                if_names.add(interface["name"])
            else:
                if_names.update(self.get_ifnames(interfaces_lst=interface))
        return if_names

    def run_as_root(self):
        """
        Restart the script as root
        """
        if os.geteuid() != 0:
            print("You need to be root to run this script. Relaunching with "
                  "sudo...\n")
            subprocess.call(["sudo", sys.executable] + sys.argv)
            exit()

    def apply_qos(self):
        self.run_as_root()
        # Clean old rules
        self.reset_qos()
        # Setting new rules
        print("Setting new rules")
        for r in self.run_list:
            r.apply(dryrun=self.config.get("DRYRUN", False))

    def reset_qos(self):
        """
        Reset QoS for all configured interfaces
        """
        self.run_as_root()
        print("Removing tc rules")
        ifnames = self.get_ifnames()
        tc.qdisc_del(ifnames, stderr=subprocess.DEVNULL)

    def show_qos(self):
        ifnames = self.get_ifnames()
        print("\n\t QDiscs details\n\t================\n")
        tc.qdisc_show(ifnames, "details")
        print("\n\t QDiscs stats\n\t==============\n")
        tc.qdisc_show(ifnames, "details")

    def init_parser(self):
        """
        Init argparse objects
        """
        parser = argparse.ArgumentParser(
            description="Tool to set, show or delete QoS rules on Linux"
        )

        # Start/Stop/Show command
        sp_action = parser.add_subparsers()
        sp_start = sp_action.add_parser("start", help="set QoS rules")
        sp_stop = sp_action.add_parser("stop", help="remove all QoS rules")
        sp_show = sp_action.add_parser("show", help="show QoS rules")

        # Set function to call for each options
        sp_start.set_defaults(func=self.apply_qos)
        sp_stop.set_defaults(func=self.reset_qos)
        sp_show.set_defaults(func=self.show_qos)

        # Debug option
        parser.add_argument('-d', '--debug', help="set the debug level",
                            dest="debug", action="store_true")
        parser.add_argument('-D', '--dryrun', help="dry run",
                            dest="dryrun", action="store_true")

        self.arg_parser = parser

    def run(self):
        self.init_parser()

        # If no argument provided show help
        if len(sys.argv) == 1:
            self.arg_parser.print_help()
            sys.exit(1)

        # Parse argument
        args = self.arg_parser.parse_args()

        self.dryrun = args.dryrun
        if args.debug or args.dryrun:
            self.debug = True

        # Execute correct function, or print usage
        if hasattr(args, "func"):
            args.func()
        else:
            self.arg_parser.print_help()
            sys.exit(1)



#!/usr/bin/env python3
# Author: Anthony Ruhier

from fcntl import ioctl
import logging
import socket
import subprocess
import struct

_logger = logging.getLogger(__name__)


def get_mtu(ifname):
    """
    Use socket ioctl call to get MTU size of an interface
    """
    SIOCGIFMTU = 0x8921
    s = socket.socket(type=socket.SOCK_DGRAM)
    ifr = ifname + '\x00'*(32-len(ifname))
    try:
        ifs = ioctl(s, SIOCGIFMTU, ifr)
        mtu = struct.unpack('<H', ifs[16:18])[0]
    except Exception:
        _logger.warning("Cannot find the MTU of %s. Will use 1500", ifname)
        mtu = 1500
    return mtu


def launch_command(command, stderr=None, dryrun=False):
    """
    If the script is launched in debug mode, just prints the command.
    Otherwise, starts it with subprocess.call()
    """
    _logger.debug(" ".join(command))
    if dryrun:
        return
    r = subprocess.call(command, stderr=stderr)
    if r != 0:
        if stderr == subprocess.DEVNULL:
            return
        _logger.error(" ".join(command))


def get_child_qdiscid(classid):
    """
    Return the id to handle for a child qdisc. By convention, it will take its
    parent class id

    :param classid: parent class id
    """
    return classid[classid.find(":") + 1:]



#!/usr/bin/env python3
# Author: Anthony Ruhier

import errno
import importlib
import logging
import os
import types


class Config(dict):
    """
    Works like a dict but can be filled directly from a python configuration
    file. Inspired from the Flask Config class (a part of their code has been
    copied here).

    Only uppercase keys are added to the config.  This makes it possible to use
    lowercase values in the config file for temporary values that are not added
    to the config or to define the config keys in the same file that implements
    the application.

    :param root_path: path to which files are read relative from.  When the
                      config object is created by the application, this is
                      the application's :attr:`~pyqos.PyQoS.root_path`.
    :param defaults: an optional dictionary of default values
    """
    def __init__(self, root_path, defaults=None):
        dict.__init__(self, defaults or {})
        self.root_path = root_path or "./"
        self.refresh_global_logger_lvl()

    def refresh_global_logger_lvl(self):
        if self["DEBUG"] or self["DRYRUN"]:
            logging.getLogger("pyqos").setLevel(logging.DEBUG)

    def from_pyfile(self, filename, silent=False):
        """
        Updates the values in the config from a Python file.  This function
        behaves as if the file was imported as module with the
        :meth:`from_object` function.

        :param filename: the filename of the config.  This can either be an
                         absolute filename or a filename relative to the
                         root path.
        :param silent: set to ``True`` if you want silent failure for missing
                       files.
        """
        filename = os.path.join(self.root_path, filename)
        d = types.ModuleType('config')
        d.__file__ = filename
        try:
            with open(filename) as config_file:
                exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        self.from_object(d)
        return True

    def from_object(self, obj):
        """
        Updates the values from the given object.  An object can be of one
        of the following two types:

        - a string: in this case the object with that name will be imported
        - an actual object reference: that object is used directly

        Objects are usually either modules or classes.
        Just the uppercase variables in that object are stored in the config.
        Example usage::

            app.config.from_object('yourapplication.default_config')
            from yourapplication import default_config
            app.config.from_object(default_config)

        You should not use this function to load the actual configuration but
        rather configuration defaults.  The actual config should be loaded
        with :meth:`from_pyfile` and ideally from a location not within the
        package because the package might be installed system wide.

        :param obj: an import name or object
        """
        if isinstance(obj, str):
            obj = importlib.import_module(obj)
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)


class ConfigAttribute(object):
    """
    Makes an attribute forward to the config

    Again, copied from the Flask project
    """
    def __init__(self, name, get_converter=None):
        self.__name__ = name
        self.get_converter = get_converter

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        rv = obj.config[self.__name__]
        if self.get_converter is not None:
            rv = self.get_converter(rv)
        return rv

    def __set__(self, obj, value):
        # When setting the log level, change the global logger level
        obj.config[self.__name__] = value
        obj.config.refresh_global_logger_lvl()



#!/usr/bin/env python3
# Author: Anthony Ruhier

"""
Framework that helps setting a QoS on Linux
See:
    https://github.com/Anthony25/python_tc_qos
"""

from os import path
from setuptools import setup

here = path.abspath(path.dirname(__file__))

setup(
    name="pyqos",
    version="0.2.1",

    description="Framework that helps setting a QoS on Linux",
    long_description=open(
        path.join(path.dirname(__file__), "README.mkd")
    ).read(),

    url="https://github.com/Anthony25/pyqos",
    author="Anthony25 <Anthony Ruhier>",
    author_email="anthony.ruhier@gmail.com",

    license="Simplified BSD",

    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: System :: Networking",
        "License :: OSI Approved :: BSD License",
    ],

    keywords="networking qos linux development",
    packages=["pyqos", ],
    install_requires=["argparse", ],
    setup_requires=['pytest-runner', ],
    tests_require=['pytest', 'pytest-cov', "pytest-mock", "pytest-xdist"],
)
